# rfc2396 (1738, 1808, 2368)
# $Id: uri.rb,v 4.27 2000/12/23 17:44:48 rgt Exp $

=begin
  URI is a abstract class. But it has common interface to instantiate
  an URI object.

  To instanciate an ALL kind of URI object, use URI.create. DO NOT use new() to
  instanciate HTTPURL or so on.

  To escape query string, use URI#escape. (Thanx to NaHi, Wakou)
=end
class URI
  SCHEME   = 'scheme'.freeze
  PATH     = 'path'.freeze
  QUERY    = 'query'.freeze
  FRAGMENT = 'fragment'.freeze
  USERINFO = 'userinfo'.freeze
  HOST     = 'host'.freeze
  PORT     = 'port'.freeze
  USER     = 'user'.freeze
  PASSWORD = 'password'.freeze
  TYPECODE = 'type'.freeze

  URI_SCHEMES = {}

  ##pre absolute_uri and ((relative_uri.nil? and
  ##                       absolute_uri.is_a? String) or
  ##                      (relative_uri.is_a? String and 
  ##                       (absolute_uri.is_a? String or
  ##                        absolute_uri.is_a? URI)))
  ##post const absolute_uri, const relative_uri
  ##return aURI
  ##raise URIError if absolute_uri and/or relative_uri are a bad URI
  def URI.create(absolute_uri, relative_uri=nil)
    if absolute_uri.is_a? String
      if (uri_class = URI_SCHEMES[absolute_uri.sub(/:.*/m, '').downcase])
	absolute_uri = uri_class.new(absolute_uri)
      else
	raise URIError, 'unknown scheme: ' + absolute_uri
      end
    end
    if relative_uri
      absolute_uri.create_relative_uri(relative_uri)
    else
      absolute_uri
    end
  end

=begin
  escape str.
  default unsafe is ^uric. see URI::Escape.pm in perl5.
  if you want to escape str for CGI, use CGI::escape.
=end
  ##pre str.is_a? String and (unsafe.nil? or unsafe.is_a? String)
  def URI.escape(str, unsafe=nil)
    if not unsafe
      str.gsub(/[^;\/?:@&=+$,0-9A-Za-z\-_.!~*'()\[\]]/) do |match|
	sprintf('%%%02X', match.unpack('C')[0])
      end
    else
      str.gsub(/[#{unsafe}]/) do |match|
	sprintf('%%%02X', match.unpack('C')[0])
      end
    end
  end

  def URI.unescape(str)
    result = str.dup
    while /%[0-9A-Fa-f][0-9A-Fa-f]/ =~ result
      char = $&[1,2].hex.chr
      result.sub!(/%[0-9A-Fa-f][0-9A-Fa-f]/, char)
    end
    result
  end

  def URI.unescape_for_cgi(str)
    result = URI.unescape(str)
    result.gsub(/\+/, ' ')
  end

  def initialize(uri)
    @components = {}
    if not (@components[SCHEME] = URI::URI_SCHEMES.index(type))
      raise URIError, 'internal error: unknown scheme'
    end
  end

  def [](component_name)
    @components[component_name]
  end

  def each(&block)
    @components.each do |key, value|
      block.call key, value
    end
  end

  def scheme
    @components[SCHEME]
  end

  def ==(obj)
    if obj.is_a? String
      begin
	obj = URI.create(obj)
      rescue URIError
	return false
      end
    end
    if type != obj.type
      return false
    end
    @components[SCHEME] == obj[SCHEME]
  end

  def dup
    result = super
    result.init(@components)
    result
  end

  ##protected
  def init(hashed_components)
    @components = {}
    hashed_components.each do |key, value|
      @components[key] = value.dup
    end
  end
  protected :init

  # escape all chars except escaped
  ##private
  def escape_all(str)
    tmp = nil
    buff = ''
    arry = str.split(//)
    i = 0
    while i < arry.size
      if arry[i] == '%'
	(tmp = arry[0, 3].join('')).upcase!
	if /%[0-9A-Z][0-9A-Z]/ =~ tmp
	  buff << tmp
	  i += 3
	  next
	else
	  buff << sprintf("%%%02X", ?%)
	end
      else
	buff << sprintf("%%%02X", arry[i][0])
      end
      i += 1
    end
    buff
  end
  private :escape_all

=begin
  GenericURI is a abstract class. Its super class is URI. Its subclasses are
  FTPURL, HTTPURL and etc.

  A instance of subclass which are subclassed from GenericURI may 
  have scheme, userinfo, host, port, path, query, and fragment.

  create_relative_uri is a instance method to create another URI instance
  relative to the receiver.
=end
  class GenericURI <URI
    def GenericURI.well_known_port  # used in ==
      '0'
    end
    
    def initialize(uri)
      super(uri)
      userinfo = host = port = path = nil
      scheme, authority, path, query, fragment = parse_generic_uri(uri)

      if authority.nil?
	raise URIError, 'undefined authority'
      end
      userinfo, host, port = parse_server(authority)
      @components[USERINFO] = userinfo if userinfo
      @components[HOST] = host
      @components[PORT] = port if port
      if not path or path == ''
	@components[PATH] = '/'
      else
	@components[PATH] = resolve_path(path)
      end
      @components[QUERY] = query if query
      @components[FRAGMENT] = fragment if fragment
    end

=begin
  parses URI string into scheme, authority, path, query, and fragment.
  scheme is not checked.
=end
    ##pre uri and uri.is_a? String
    ##post const uri
    ##return [scheme, authority, path, query, fragment]
    ##raise URIError if uri is a bad URI
    ##private
    def parse_generic_uri(uri)
      x_QUERY_FRAGMENT =
	/\A(?:[0-9A-Za-z\-_.!~*'();\/?:@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*\z/

      scheme, authority, path, query, fragment = get_generic_components(uri)

      check_authority authority
      check_path path
      if query and x_QUERY_FRAGMENT !~ query
	raise URIError, 'bad query'
      end
      if fragment and x_QUERY_FRAGMENT !~ fragment
	raise URIError, 'bad fragment'
      end
      if path == '' and (query or fragment)
	path = nil
      end
      if query == '' and fragment
	query = nil
      end
      [scheme, authority, path, query, fragment]
    end
    private :parse_generic_uri

    def get_generic_components(uri)
      %r{\A(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?\z} =~ uri
      #    12            3  4          5       6  7        8 9
      # scheme = $2, authority = $4, path = $5, query = $7, fragment = $9
      [$2, $4, $5, $7, $9]
    end
    private :get_generic_components

    def check_authority(authority)
      return if authority.nil?
      # IPv6-authority is not checked. it is checked by parse_server
      return if /\[.*\]/ =~ authority
      if /\A(?:[0-9A-Za-z\-_.!~*'()$,;:@&=+]|%[0-9A-Fa-f][0-9A-Fa-f])+\z/ !~
	  authority
	raise URIError, 'bad authority' + authority
      end
    end
    private :check_authority

    # if uri starts with rel_segment, make it '/rel_segment' temporarily.
    # a bad rel_segment, which contains ':', does not reach here.
    # even if uri contains abs_path, '/abs_path' is valid abs_path.
    # see RFC's BNF
    def check_path(path)
      return if path.nil?
      if /\A(?:(?:\/)(?:(?:[0-9A-Za-z\-_.!~*'():@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*(?:;(?:[0-9A-Za-z\-_.!~*'():@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*)*))+\z/ !~
	  ('/' << path)
	raise URIError, 'bad path'
      end
    end
    private :check_path

    # userinfo@host:port
    ##private
    def parse_server(authority)
      userinfo, host, port = parse_IPv6_server(authority)

      if host.nil?
	userinfo, host, port = parse_IPv4_server(authority)
      end
      if userinfo and
	  /\A(?:[0-9A-Za-z\-_.!~*'();:&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*\z/ !~
	  userinfo
	raise URIError, 'bad userinfo'
      end
      if port and /\A\d*\z/ !~ port
	raise URIError, 'bad port'
      end
      [userinfo, host, port]
    end
    private :parse_server

    # see RFC-2373 and RFC-2732
    ##private
    def parse_IPv6_server(authority)
      mESSAGE = 'bad IPv6 address'.freeze

      return nil if /\A(?:([^@]*)@)?(\[(.+)\])(?::(.*))?\z/ !~ authority
      userinfo = $1
      host = $2
      addr = $3
      port = $4

      if addr.size < 2 or
	  (addr[0] == ?: and addr[1] != ?:) or
	  (addr[-1] == ?: and addr[-2] != ?:) or
	  addr.index ':::'
	raise URIError, mESSAGE
      end

      check_IPv6_double_colon addr
      check_IPv6_hex addr
      [userinfo, host, port]
    end
    private :parse_IPv6_server

    ##private
    def check_IPv6_double_colon(addr)
      mESSAGE = 'bad IPv6 address'.freeze

      doubles = addr.scan(/::/).size
      raise URIError, mESSAGE if doubles > 1

      hexes = addr.split(':').size
      if (doubles == 1 and hexes > 6) or
	  (doubles == 0 and hexes > 8)
	  raise URIError, mESSAGE
      end
    end
    private :check_IPv6_double_colon

    ##private
    def check_IPv6_hex(addr)
      mESSAGE = 'bad IPv6 address'.freeze

      arry = addr.split(':')
      remove_IPv4_addr arry
      arry.each do |hex|
	raise URIError, mESSAGE if /\A[0-9A-Fa-f]{0,4}\z/ !~ hex
      end
    end
    private :check_IPv6_hex

    ##private
    def remove_IPv4_addr(addr_arry)
      if /\A\d+(?:\.\d+){3}\z/ =~ addr_arry[-1]
	addr_arry.pop
      end
    end
    private :remove_IPv4_addr

    ##private
    def parse_IPv4_server(authority)
      userinfo, host, port =
	authority.scan(/\A(?:([^@]*)@)?([^:]*)(?::(.*))?\z/)[0]
      if not host or
	  /\A(?:(?:(?:(?:(?:[0-9A-Za-z]|[0-9A-Za-z][0-9A-Za-z\-]*[0-9A-Za-z])(?:\.))*)(?:[A-Za-z]|[A-Za-z][0-9A-Za-z\-]*[0-9A-Za-z])(?:\.?))|(?:\d+\.\d+\.\d+\.\d+))\z/ !~ host
	raise URIError, 'bad host'
      end
      [userinfo, host, port]
    end
    private :parse_IPv4_server

    # resolve absolute path
    # see RFC-2396, C.2. Abnormal Examples
    ##private
    def resolve_path(path)
      if path == '/./' or path == '/.' or path == '/../' or path == '/..'
	return path.dup
      end
      result = remove_current_dir path
      remove_parent_dir result
    end
    private :resolve_path

    ##private
    def remove_current_dir(path)
      sLASH = '/'.freeze

      result = path.dup
      if %r|/\.\.?\z| =~ path  #/
	result << sLASH
      end

      result.gsub!(%r|/(\./)+|, sLASH)
      if path[0] == ?/ and path[1] == ?. and path[2] == ?/
	  result = '/.' << result
      end
      result
    end
    private :remove_current_dir

    ##private
    def remove_parent_dir(apath)
      dOT_DOT = '..'.freeze
      result = []
      apath.split(%r|/|, -1).each do |elt|  #/
	if elt == dOT_DOT
	  if result.size == 1 or result[-1] == dOT_DOT
	    result.push elt
	  else
	    result.pop
	  end
	else
	  result.push elt
	end
      end
      result.join('/')
    end
    private :remove_parent_dir

    def create_relative_uri(relative_uri)
      buff = nil
      scheme, authority, path, query, fragment =
	parse_generic_uri(relative_uri)
      if scheme
	raise URIError, 'bad relative uri'
      end
      if authority
	return type.new(relative_uri)
      end
      buff = @components[SCHEME] + '://'
      buff << @components[USERINFO] << '@' if @components[USERINFO]
      buff << @components[HOST]
      buff << ':' << @components[PORT] if @components[PORT]
      if path
	return create_relative_uri_with_path(path, buff, relative_uri)
      end
      buff << @components[PATH]
      if query
	buff.sub!(%r|[^/]*\z|, relative_uri)  #/
	return type.new(buff)
      end
      if @components[QUERY]
	buff << '?' << @components[QUERY]
      end
      if fragment
	return type.new(buff << relative_uri)
      end
      return type.new(buff)
    end

    ##private
    def create_relative_uri_with_path(path, buff, relative_uri)
      if path[0] == (?/)  #/
	type.new(buff << relative_uri)
      else
	type.new(buff << @components[PATH].sub(%r|[^/]*\z|, relative_uri))  #/
      end
    end
    private :create_relative_uri_with_path

    def host
      @components[HOST]
    end
    def host=(str)
      @components[HOST] = str
    end
    def path
      @components[PATH]
    end
    def path=(str)
      @components[PATH] = str
    end
    def port
      if @components[PORT]
	@components[PORT]
      else
	type.well_known_port
      end
    end
    def port=(port)
      raise URIError, 'digits required' if /^\d+$/ !~ port.to_s
      @components[PORT] = port.to_s
    end
    def query
      @components[QUERY]
    end
    def query=(str)
      @components[QUERY] = str
    end
    def fragment
      @components[FRAGMENT]
    end
    def fragment=(str)
      @components[FRAGMENT] = str
    end
    def userinfo
      @components[USERINFO]
    end
    def userinfo=(str)
      @components[USERINFO] = str
    end

    def ==(obj)
      m = y = nil  # mine, yours
      if obj.is_a? String
	begin
	  obj = URI.create(obj)
	rescue URIError
	  return false
	end
      end
      if not super(obj) or
	  not_equal_userinfo? obj or
	  not_equal_host? obj or
	  port != obj.port or
	  not_equal_path? obj or
	  not_equal_query? obj or
	  not_equal_fragment? obj
	return false
      end
      true
    end

    ##private
    def not_equal_userinfo?(other)
      if @components[USERINFO] and other.userinfo and
	  @components[USERINFO] != other.userinfo
	return true
      elsif @components[USERINFO] or other.userinfo
	return true
      end
    end
    private :not_equal_userinfo?

    ##private
    def not_equal_host?(other)
      if @components[HOST] == other.host or
	  escape_all(@components[HOST].downcase) ==
	  escape_all(other.host.downcase)
	return false
      end
      true
    end
    private :not_equal_host?

    ##private
    def not_equal_query?(other)
      return false if @components[QUERY] == other.query
      if @components[QUERY] and other.query and
	  escape_all(@components[QUERY]) == escape_all(other.query)
	return false
      end
      true
    end
    private :not_equal_query?

    ##private
    def not_equal_fragment?(other)
      return false if @components[FRAGMENT] == other.fragment
      if @components[FRAGMENT] and other.fragment and
	  escape_all(@components[FRAGMENT]) == escape_all(other.fragment)
	return false
      end
      true
    end
    private :not_equal_fragment?

    ##private
    def not_equal_path?(other)
      return false if @components[PATH] == other.path

      arry_a = @components[PATH].split('/', -1)
      arry_b = other.path.split('/', -1)
      return true if arry_a.size != arry_b.size

      arry_a.size.times do |i|
	if arry_a[i] != arry_b[i] and
	    escape_all(arry_a[i]) != escape_all(arry_b[i])
	  return true
	end
      end
      false
    end
    private :not_equal_path?

    def to_s
      tmp = nil
      buff = @components[SCHEME] + '://'
      if @components[USERINFO]
	buff << @components[USERINFO] << '@'
      end
      buff << @components[HOST]
      if @components[PORT]
	buff << ':' << @components[PORT]
      end
      buff << @components[PATH]
      if @components[QUERY]
	buff << '?' << @components[QUERY]
      end
      if @components[FRAGMENT]
	buff << '#' << @components[FRAGMENT]
      end
      buff
    end
  end

=begin
  To instanciate a FTPURL object, use URI.create.
  A FTPURL object has scheme, userinfo (user, password), host, port, path, 
  and typecode.
=end
  class FTPURL <GenericURI
    URI::URI_SCHEMES['ftp'.freeze] = self

    WELL_KNOWN_PORT = '21'.freeze

    def FTPURL.well_known_port  # used in GenericURI#==
      FTPURL::WELL_KNOWN_PORT
    end

    def initialize(absolute_uri)
      super(absolute_uri)
      if @components[QUERY] or @components[FRAGMENT]
	raise URIError, 'bad FTP URL'
      end

      init_userinfo
      init_typecode
    end

    def init_userinfo
      return if @components[USERINFO].nil?
      if @components[USERINFO].index(':').nil?
	raise URIError, 'bad userinfo'
      end

      @components[USER], @components[PASSWORD] =
	@components[USERINFO].split(':', 2)
    end
    private :init_userinfo

    def init_typecode
      path, typecode = @components[PATH].split(';')

      return if typecode.nil?
      if /\Atype=([aid])\z/ !~ typecode
	raise URIError, 'bad type'
      end

      @components[TYPECODE] = $1
    end
    private :init_typecode
    
    def user
      @components[USER]
    end
    def user=(str)
      @components[USER] = str
    end
    def password
      @components[PASSWORD]
    end
    def password=(str)
      @components[PASSWORD] = str
    end
    def typecode
      @components[TYPECODE]
    end
    def typecode=(typecode)
      raise URIError, 'illegal typecode' if /^[aid]$/ !~ typecode  #/
      @components[TYPECODE] = str
    end
  end

=begin
  To instanciate a HTTPURL object, use URI.create.
  a HTTPURL object has scheme, host, port, path, query, and fragment.
=end
  class HTTPURL <GenericURI
    URI::URI_SCHEMES['http'.freeze] = self

    WELL_KNOWN_PORT = '80'.freeze

    def HTTPURL.well_known_port  # used in GenericURI#==
      HTTPURL::WELL_KNOWN_PORT
    end

    def initialize(absolute_uri)
      super(absolute_uri)
      if @components[USERINFO]
	raise URIError, 'bad authority'
      end
    end
  end
end

class URIError < StandardError
end
