# urireg.rb:  generates Regexps for uri.rb
# $Id: urireg.rb,v 1.3 2000/01/13 22:15:16 rgt Exp $

# CAUTION: Some regexps generated by this cannot be compiled.
#          In ruby v1.4.3, /\//.source() generates "/".

=begin
rfc2396
      URI-reference = [ absoluteURI | relativeURI ] [ "#" fragment ]
      absoluteURI   = scheme ":" ( hier_part | opaque_part )
      relativeURI   = ( net_path | abs_path | rel_path ) [ "?" query ]

      hier_part     = ( net_path | abs_path ) [ "?" query ]
      opaque_part   = uric_no_slash *uric

      uric_no_slash = unreserved | escaped | ";" | "?" | ":" | "@" |
                      "&" | "=" | "+" | "$" | ","

      net_path      = "//" authority [ abs_path ]
      abs_path      = "/"  path_segments
      rel_path      = rel_segment [ abs_path ]

      rel_segment   = 1*( unreserved | escaped |
                          ";" | "@" | "&" | "=" | "+" | "$" | "," )

      scheme        = alpha *( alpha | digit | "+" | "-" | "." )

      authority     = server | reg_name

      reg_name      = 1*( unreserved | escaped | "$" | "," |
                          ";" | ":" | "@" | "&" | "=" | "+" )

      server        = [ [ userinfo "@" ] hostport ]
      userinfo      = *( unreserved | escaped |
                         ";" | ":" | "&" | "=" | "+" | "$" | "," )

      hostport      = host [ ":" port ]
      host          = hostname | IPv4address
      hostname      = *( domainlabel "." ) toplabel [ "." ]
      domainlabel   = alphanum | alphanum *( alphanum | "-" ) alphanum
      toplabel      = alpha | alpha *( alphanum | "-" ) alphanum
      IPv4address   = 1*digit "." 1*digit "." 1*digit "." 1*digit
      port          = *digit

      path          = [ abs_path | opaque_part ]  # path is dangling
      path_segments = segment *( "/" segment )
      segment       = *pchar *( ";" param )
      param         = *pchar
      pchar         = unreserved | escaped |
                      ":" | "@" | "&" | "=" | "+" | "$" | ","

      query         = *uric

      fragment      = *uric

      uric          = reserved | unreserved | escaped
      reserved      = ";" | "/" | "?" | ":" | "@" | "&" | "=" | "+" |
                      "$" | ","
      unreserved    = alphanum | mark
      mark          = "-" | "_" | "." | "!" | "~" | "*" | "'" |
                      "(" | ")"

      escaped       = "%" hex hex
      hex           = digit | "A" | "B" | "C" | "D" | "E" | "F" |
                              "a" | "b" | "c" | "d" | "e" | "f"

      alphanum      = alpha | digit
      alpha         = lowalpha | upalpha

      lowalpha = "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h" | "i" |
                 "j" | "k" | "l" | "m" | "n" | "o" | "p" | "q" | "r" |
                 "s" | "t" | "u" | "v" | "w" | "x" | "y" | "z"
      upalpha  = "A" | "B" | "C" | "D" | "E" | "F" | "G" | "H" | "I" |
                 "J" | "K" | "L" | "M" | "N" | "O" | "P" | "Q" | "R" |
                 "S" | "T" | "U" | "V" | "W" | "X" | "Y" | "Z"
      digit    = "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" |
                 "8" | "9"
=end

require 'regman'

# pchar = unreserved | escaped | ":" | "@" | "&" | "=" | "+" | "$" | ","
PCHAR = /[0-9A-Za-z\-_.!~*'():@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f]/

# param = *pchar
PARAM = /(?:[0-9A-Za-z\-_.!~*'():@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*/

# segment = *pchar *( ";" param )
SEGMENT = /(?:[0-9A-Za-z\-_.!~*'():@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*(?:;(?:[0-9A-Za-z\-_.!~*'():@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*)*/

# IPv4address = 1*digit "." 1*digit "." 1*digit "." 1*digit
IPV4ADDRESS = /\d+\.\d+\.\d+\.\d+/

# toplabel = alpha | alpha *( alphanum | "-" ) alphanum
TOPLABEL = /[A-Za-z]|[A-Za-z][0-9A-Za-z\-]*[0-9A-Za-z]/

# domainlabel = alphanum | alphanum *( alphanum | "-" ) alphanum
DOMAINLABEL = /[0-9A-Za-z]|[0-9A-Za-z][0-9A-Za-z\-]*[0-9A-Za-z]/

# hostname = *( domainlabel "." ) toplabel [ "." ]
HOSTNAME = RM.cat(RM.star(RM.cat(DOMAINLABEL, /\./)), TOPLABEL, /\.?/)

# path_segments = segment *( "/" segment )
tmp = RM.cat(SEGMENT, RM.star(RM.cat(/\//, SEGMENT)))
PATH_SEGMENTS = /\A#{tmp.source}\z/

# abs_path = "/" path_segments
#  equals
# abs_path = "/" segment *( "/" segment )
#  equals
# abs_path = 1*( "/" segment )
tmp = RM.plus(RM.cat(/\//, SEGMENT))
ABS_PATH = /\A#{tmp.source}\z/

# rel_segment = 1*( unreserved | escaped |
#                   ";" | "@" | "&" | "=" | "+" | "$" | "," )
REL_SEGMENT = /\A(?:[0-9A-Za-z\-_.!~*'();@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])+\z/

# host = hostname | IPv4address
tmp = RM.bar(HOSTNAME, /\d+\.\d+\.\d+\.\d+/)
HOST = /\A(?:#{tmp.source})\z/

# userinfo = *( unreserved | escaped |
#               ";" | ":" | "&" | "=" | "+" | "$" | "," )
USERINFO = /\A(?:[0-9A-Za-z\-_.!~*'();:&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*\z/

# reg_name = 1*( unreserved | escaped | "$" | "," |
#                ";" | ":" | "@" | "&" | "=" | "+" )
REG_NAME = /\A(?:[0-9A-Za-z\-_.!~*'()$,;:@&=+]|%[0-9A-Fa-f][0-9A-Fa-f])+\z/

# query = *uric
# uric  = reserved | unreserved | escaped
QUERY = /\A(?:[0-9A-Za-z\-_.!~*'();\/?:@&=+$,]|%[0-9A-Fa-f][0-9A-Fa-f])*\z/

# scheme = alpha *( alpha | digit | "+" | "-" | "." )
SCHEME = /\A[A-Za-z][0-9A-Za-z+\-.]*\z/

# fragment = *uric
FRAGMENT = QUERY

puts 'FRAGMENT (invalid source)'
puts FRAGMENT.source
puts

puts 'SCHEME'
puts SCHEME.source
puts

puts 'QUERY (invalid source)'
puts QUERY.source
puts

puts 'REG_NAME'
puts REG_NAME.source
puts

puts 'USERINFO'
puts USERINFO.source
puts

puts 'HOST'
puts HOST.source
puts

puts 'REL_SEGMENT'
puts REL_SEGMENT.source
puts

puts 'PATH_SEGMENTS (invalid source)'
puts PATH_SEGMENTS.source
puts

puts 'ABS_PATH (invalid source)'
puts ABS_PATH.source
