# $Id: test-uri.rb,v 5.11 2000/10/03 21:47:57 rgt Exp $

require 'runit/testcase'
require 'runit/cui/testrunner'
require 'uri'

include RUNIT

class URITester <TestCase
  INPUT_FILE = 'in-uri.txt'

  BASE = 'base:'.freeze
  ABSOLUTE = 'absolute:'.freeze
  RELATIVE = 'relative:'.freeze

  def setup
    @lines = IO.readlines(INPUT_FILE)
    @lines.delete_if { |line| line[0] == ?# or line == $/ }
  end

  def test_all
    base = current = nil
    @lines.each do |line|
      type, input, answer = line.split
      if answer.nil?
	answer = input
      end
      type.strip!
      input.strip!
      answer.strip!
      begin
	case type
	when BASE
	  base = current = URI.create(input)
	when RELATIVE
	  current = base.create_relative_uri(input)
	when ABSOLUTE
	  current = URI.create(input)
	else
	  raise 'unknown type'
	end
      rescue URIError
	printf '%s: %s%s', $., $!, ($/)
      end
      if current.to_s != answer
	printf 'ERROR:%s: %s <=> %s%s', $., current.to_s, answer, ($/)
      end
    end
  end

  def test_create
    uri = URI.create('http://www02.so-net.ne.jp/~greentea/')
    assert(uri.is_a? URI::HTTPURL)
  end
  
  def test_setters
    expected = 'http://www.ruby-lang.org:80/en/whos.html'
    uri = URI.create('http://www02.so-net.ne.jp/~greentea/')
    uri.host = 'www.ruby-lang.org'
    uri.port = 80
    uri.path = '/en/whos.html'
    assert_equal expected, uri.to_s
  end

  def test_escape
    uric = make_uric
    buff = ''
    1.upto 255 do |c|
      str = URI.escape(c.chr)
      if str.size == 1
	buff << str
      end
    end

    assert_equal uric, buff
  end

  def make_uric
    # uric          = reserved | unreserved | escaped
    # reserved      = ";" | "/" | "?" | ":" | "@" | "&" | "=" | "+" |
    #                "$" | "," | "[" | "]"
    # unreserved    = alphanum | mark
    # mark          = "-" | "_" | "." | "!" | "~" | "*" | "'" |
    #                 "(" | ")"

    result = ";" << "/" << "?" << ":" << "@" << "&" << "=" << "+" << "$" <<
      "," << "[" << "]" << "-" << "_" << "." << "!" << "~" << "*" << "'" <<
      "(" << ")" #'
    0.upto 9 do |i| result << i.to_s end
    ?A.upto ?Z do |a| result << a.chr end
    ?a.upto ?z do |a| result << a.chr end
    result.split(//).sort.join
  end
  private :make_uric

  def test_equal
    a = URI.create('http://www02.so-net.ne.jp/~greentea/index.html?a#b')
    b = URI.create('http://www02.So-Net.ne.jp:80/~greentea/index.html?%61#%62')
    assert_equal a, b
  end

  def test_FTPURL
    expected = 'ftp://greentea:foobar@ftp.ruby-lang.org/ruby;type=a'
    url = URI.create(expected)
    assert url.is_a? URI::FTPURL
    assert_equal 'greentea', url.user
    assert_equal 'foobar', url.password
    assert_equal '/ruby;type=a', url.path
    assert_equal 'a', url.typecode
    assert_equal expected, url.to_s
  end

  def test_escape_for_cgi
    expected = '~a b'

    assert_equal expected, URI.unescape_for_cgi '%7Ea+b'
  end
end

CUI::TestRunner.run URITester.suite
